import csv
from .argument import check_argument
from .converter.func import convert_table


@convert_table
@check_argument
def load_csv(
        file_path,
        encoding='utf-8-sig'):

    table = []
    with open(file_path, encoding=encoding) as fd:
        for record in csv.DictReader(fd):
            table.append(record)

    return table



from functools import wraps
from ..func_mode import get_func_mode
from .converters import CONVERTERS
from .default import default_load
from .default import default_dump


def convert_table(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        mode = get_func_mode(func)

        if mode == 'load':
            converters = kwargs.pop('converters', default_load)

            table = func(*args, **kwargs)

            table = apply_converters(table, converters)

            return table

        elif mode == 'dump':
            converters = kwargs.pop('converters', default_dump)

            # TODO, maybe check which position contains the table in parameter
            table = args[1]
            table = apply_converters(table, converters)

            args = list(args)
            args[1] = table

            return func(*args, **kwargs)

    return wrapper


def apply_converters(table, converters):

    # print(converters)
    for c in converters:
        arg = None
        if type(c) in (list, tuple):
            c, arg = c
        if type(c) == str:
            c = CONVERTERS.get(c)

        if not c:
            raise Exception(f'Converter {c} not found')

        if arg:
            table = c(table, arg)
        else:
            table = c(table)

    return table



from ..levels import CellConverter
from ..levels import RowConverter
from ..levels import ColumnConverter
from ..levels import TableConverter
import pkgutil

# __all__ = []


def get_level_converter(level):
    if level == 'row':
        return RowConverter
    elif level == 'cell':
        return CellConverter
    elif level == 'column':
        return ColumnConverter
    else:
        return TableConverter


CONVERTERS = {}


for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    # __all__.append(module_name)

    module = loader.find_module(
        module_name).load_module(module_name)

    if not hasattr(module, 'level'):
        continue

    converter = get_level_converter(module.level)

    CONVERTERS[module_name] = converter(
        name=module_name,
        asserter=module.asserter,
        converter=module.converter
    )


# TODO: Column version
# TODO:
# 1. keys are customized functions
# 2. design an easy way to direct how to sort

# TODO, level show in the module file name, not forced to do that, because that will give less choice to name the function



BOOLEAN_TABLE = {
    True: ('true', 'yes', '1'),
    False: ('false', 'no', '0'),
}


level = 'column'
asserter = lambda x: type(x) == str
# TODO should test all cell in the column
converter = lambda x: {
            i: k
            for k, v in BOOLEAN_TABLE.items()
            for i in v
        }.get(x, x)



from functools import wraps
from pathlib import Path
from .func_mode import get_func_mode
import warnings


def check_argument(func):

    @wraps(func)
    def pack(*args, **kwargs):

        func_mode = get_func_mode(func)

        # TODO get assigned by name, using functools to inspect
        # TODO: use some tricks to speed up the process
        args = list(args)

        args = check_path(args, func_mode)
        args = check_table(args, func_mode)
        kwargs = check_header(args, kwargs, func_mode)

        return func(*args, **kwargs)

    return pack


def check_table(args, func_mode):

    if func_mode != 'dump':
        return args

    file_path = args[0]
    table = args[1]
    if not table:
        warnings.warn(f'{file_path} table is empty.')

    return args


def check_header(args, kwargs, func_mode):
    if func_mode != 'dump':
        return kwargs

    # use kw dict to pass args and kwargs in check functions
    table = args[1]
    header = kwargs.get('header')

    if header:
        return kwargs

    header = []
    for rec in table:
        for key in rec.keys():
            if key not in header:
                header.append(key)

    kwargs['header'] = header

    return kwargs


# TODO: get table columns by header, using the partial info and using table converter
# def dump_csv(file_path, table, headers=[], remain=True):

#     table_headers = []
#     for rec in table:
#         for key in rec.keys():
#             if key not in table_headers:
#                 table_headers.append(key)

#     if not headers:
#         headers = table_headers
#     else:
#         remain_headers = [
#             i
#             for i in table_headers
#             if i not in headers
#         ]
#         if remain:
#             headers = headers + remain_headers
#         table = [
#             {
#                 k: v
#                 for k, v in i.items()
#                 if k in headers
#             }
#             for i in table
#         ]

#     with open(file_path, 'w', encoding='utf-8-sig') as fd:
#         writer = csv.DictWriter(fd, fieldnames=headers)
#         writer.writeheader()
#         writer.writerows(table)


def check_path(args, func_mode):

    file_path = args[0]

    if type(file_path) == str:
        file_path = Path(file_path).expanduser().resolve()

    if func_mode == 'load':
        # TODO: load file default two ways,
        #   1 raise exception,
        #   2 warning return []
        # implement a switch
        if not file_path.exists():
            raise Exception(f'{file_path} not found')
    else:
        # TODO: help use make folder or not
        file_path.parent.mkdir(exist_ok=True, parents=True)

    args[0] = file_path

    return args



import csv
from .argument import check_argument
from .converter.func import convert_table


@convert_table
@check_argument
def dump_csv(file_path, table, header=[]):
    with open(file_path, 'w', encoding='utf-8-sig') as fd:
        writer = csv.DictWriter(fd, fieldnames=header)
        writer.writeheader()
        writer.writerows(table)



def sort_table(table, keys=[]):
    if not keys:
        return

    def sorter(row):
        result = []
        for key in keys:
            result.append(row[key])
        return tuple(result)

    return sorted(table, key=sorter)


level = 'table'
asserter = True
converter = sort_table



def get_func_mode(func):

    func_name = func.__name__

    if '_' not in func_name:
        func_mode = None
    else:
        func_mode = func_name.split('_')[0]

    assert func_mode in ('load', 'dump'), f'{func_mode} not accepted'

    return func_mode



# TODO: may be us yaml to config
#   default.yaml
# config_path = Path(__file__).parent / 'converter.yml'

# for mode, c_names in load_yaml(config_path).items():
#     if mode == 'load':
#         DEFAULT_LOAD_CONVERTERS = [
#             [
#                 i
#                 for i in CONVERTER_LIST
#                 if i.name in c_names
#             ][0]
#             for c in c_names
#         ]
#     else:
#         DEFAULT_DUMP_CONVERTERS = [
#             [
#                 i
#                 for i in CONVERTER_LIST
#                 if i.name in c_names
#             ][0]
#             for c in c_names
#         ]

from .converters import CONVERTERS
from .levels import ConverterList


DEFAULT_LOAD_CONVERTERS = [
    'trim_string_by_cell',
    'remove_blank_row',
    'to_integer_cell',
    # 'boolean_from_str', TODO for cell
]
default_load = ConverterList()
for i in DEFAULT_LOAD_CONVERTERS:
    default_load.append(CONVERTERS[i])


DEFAULT_DUMP_CONVERTERS = [
    'trim_string_by_cell',
    'remove_blank_row',
    'to_null_str',
    'boolean_to_str',
    'replace_excel_return',
]
default_dump = ConverterList()
for i in DEFAULT_DUMP_CONVERTERS:
    default_dump.append(CONVERTERS[i])



from .load_csv import load_csv
from .dump_csv import dump_csv
from .converter.converters import CONVERTERS
from .converter.default import default_load
from .converter.default import default_dump


__all__ = [
    'load_csv',
    'dump_csv',
    'CONVERTERS',
    'default_load',
    'default_dump'
]



class Converter:

    TYPE = None

    def __init__(self, name, asserter, converter):
        self.name = name
        self.asserter = asserter
        self.converter = converter

    def __str__(self):

        return f'{self.__class__}({self.name})'


class ConverterList(list):

    def reinit(self):

        if len(self) > 0:
            self.pop()


class CellConverter(Converter):

    TYPE = 'cell'

    def __call__(self, table):
        new_table = []

        for rec in table:
            for k, v in rec.items():
                if self.asserter(v):
                    rec[k] = self.converter(v)

            new_table.append(rec)

        return new_table


class RowConverter(Converter):

    TYPE = 'row'

    def __call__(self, table):
        new_table = []
        for rec in table:

            if self.asserter(rec):
                rec = self.converter(rec)
            if not rec:
                continue

            new_table.append(rec)
        return new_table


class ColumnConverter(Converter):

    TYPE = 'col'

    def __call__(self, table, column_name):
        new_table = []
        for rec in table:
            val = rec[column_name]
            if self.asserter(val):
                rec[column_name] = self.converter(val)
            new_table.append(rec)

        return new_table


class TableConverter(Converter):

    TYPE = 'table'

    def __call__(self, table):
        return self.converter(table)


# TODo
# use function to collect all local converter
# the convert function can't be a configurable yaml obj, that's the limit.
# definitely for some, need an assert
